/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#ifndef PROB_H
#define PROB_H

#include <math.h>

#include "dict.h"
#include "count.h"

typedef struct _prob {
  int m, n;  /* m is size of event, n size of event + context */
  int flags[MAX_DEPTH];
  char *names[MAX_DEPTH];
  int mapping[MAX_DEPTH];
  count *nct, *uct; /* events, contexts */
  count *udct; /* contexts and diversity for lambdas */

  int cache_size;
  double *cache_p;
  int **cache_ks;

  int hits, lookups;

  struct _prob *next;
} *prob;

prob new_prob(int m, int n, int *flags, char **names);
void prob_delete(prob p);
int prob_set_cache(prob p, int cache_size);
int prob_read(prob p, const char *prefix);
int prob_read_lambdas(prob p, const char *prefix);
int prob_write(prob p, const char *prefix);
void prob_close(prob p);
int prob_set_mapping(prob p, char *names[], int n);
int prob_count(prob p, int *ks, double delta);
int prob_mapped_count(prob p, int *ks, double delta);
double prob_lookup(prob p, int *ks);
double prob_mapped_lookup(prob p, int *ks);

double prob_cache_load(prob p);
double prob_cache_hitrate(prob p);
void prob_stats(prob p);

#define IMPOSSIBLE -HUGE_VAL

#define SMALL 1e-100

#endif 
