/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdlib.h>

#include "hash_map.h"
#include "list.h"
#include "generic.h"

hash_map new_hash_map(_COMPARE(compare), _HASH(hash)) {
  hash_map ht = malloc(sizeof (*ht));

  ht->size = HASH_INITIAL_SIZE;
  ht->bucket = calloc(ht->size, sizeof(list_node));
  ht->compare = compare;
  ht->hash = hash;

  ht->nelem = 0;

  ht->accounting = 0;
  ht->lookups = 0;
  ht->lookup_time = 0;

  return ht;
}

void hash_map_delete(hash_map ht) {
  int i;
  list_node li;
  for (i=0; i<ht->size; i++) {
    for (li = ht->bucket[i]; li != NULL; li = li->next)
      free(li->contents);
    list_delete(ht->bucket[i]);
  }
  free(ht->bucket);
  free(ht);
}

int hash_map_rehash(hash_map ht, int size) {
  list_node *old_bucket = ht->bucket;
  list_node cur, next;
  int old_size = ht->size;
  int i;
  unsigned int hash;

  ht->bucket = calloc(size, sizeof(list_node));
  ht->size = size;

  for (i=0; i<old_size; i++) {
    for (cur = old_bucket[i]; cur != NULL; cur = next) {
      next = cur->next;
      hash = (*ht->hash)(((struct hash_map_entry *)cur->contents)->key) % ht->size;
      cur->next = ht->bucket[hash];
      ht->bucket[hash] = cur;
    }
  }

  free(old_bucket);
  return 0;
}

void ** hash_map_lookup(hash_map ht, void *key, int create) {
  unsigned int hash = (*ht->hash)(key) % ht->size;
  list_node *pcur;
  struct hash_map_entry *pe;
  
  if (ht->accounting && ht->lookups % 1000 == 999) {
    fprintf(stderr, "hash_map %x: average list length %f\n", ht, (float)ht->lookup_time/ht->lookups);
  }

  ht->lookups++;

  for (pcur = &ht->bucket[hash]; *pcur != NULL; pcur = &(*pcur)->next) {
    ht->lookup_time++;
    if ((*ht->compare)(key, ((struct hash_map_entry *)(*pcur)->contents)->key) == 0)
      return &((struct hash_map_entry *)(*pcur)->contents)->value;
  }

  if (create) {
    pe = malloc(sizeof(struct hash_map_entry));
    pe->key = key;
    pe->value = NULL;
    *pcur = new_list_node(pe, NULL);

    ht->nelem++;
    if (ht->nelem > ht->size * HASH_LOAD_FACTOR)
      hash_map_rehash(ht, ht->size*2+1);

    return &pe->value;
  } else
    return NULL;
}

list_node hash_map_similar(hash_map ht, unsigned int hash) {
  return ht->bucket[hash % ht->size];
}
