#include "elem_label.h"
#include "elem_node.h"
#include "dict.h"

extern dict d;

extern int show_details;

elem_label new_elem_label(void) {
  elem_label sn;
  sn = malloc(sizeof(*sn));

  sn->label = sn->headword = 0;
  sn->pge = NULL;
  sn->type = sn->spine = sn->noadj = sn->empty = sn->anchor = sn->virtual = 0;
  sn->pp = 0.0;

  return sn;
}

void elem_label_delete(elem_label sn) {
  free (sn);
}

elem_label elem_label_copy(elem_label sn) {
  elem_label nn = new_elem_label();

  if (sn == NULL)
    return NULL;

  nn->label = sn->label;

  /* This stuff gets replaced by tree_instantiate anyway, but we'll copy 
     it for completeness */
  nn->pge = sn->pge;
  nn->oheadword = sn->oheadword;
  nn->headword = sn->headword;
  nn->headtag = sn->headtag;
  nn->d_addr = sn->d_addr;
  nn->d_type = sn->d_type;
  nn->spine = sn->spine;
  nn->noadj = sn->noadj;
  nn->pp = sn->pp;

  nn->type = sn->type;
  nn->empty = sn->empty;
  nn->anchor = sn->anchor;
  nn->virtual = sn->virtual;

  return nn;
}

elem_label elem_nonterm_from_string(const char *s) {
  int len;
  char buf[80];
  elem_label sn = new_elem_label();

  strncpy(buf, s, sizeof(buf));

  len = strlen(buf);

  sn->type = INTERIOR;
  sn->label = dict_insert(d, buf);

  if (!strcmp(buf, "<"))
    sn->virtual = LEFT;
  else if (!strcmp(buf, ">"))
    sn->virtual = RIGHT;

  return sn;
}

elem_label elem_term_from_string(const char *s) {
  int len;
  char buf[80];
  elem_label sn = new_elem_label();

  strncpy(buf, s, sizeof(buf));

  len = strlen(buf);
  if (len > 1 && buf[len-1] == '*' && buf[0] != '*') {
    sn->type = FOOT;
    buf[len-1] = '\0';
    sn->label = dict_insert(d, buf);
  } else if (len > 1 && buf[len-1] == '+') {
    sn->type = SUBST;
    buf[len-1] = '\0';
    sn->label = dict_insert(d, buf);
  } else if (len > 1 && buf[len-1] == '#') {
    sn->type = FRONTIER;
    sn->empty = 1;
    buf[len-1] = '\0';
    sn->label = dict_insert(d, buf);
  } else if (len == 1 && buf[0] == '@') {
    sn->type = INTERIOR;
    sn->anchor = 1;
  } else {
    /* Fixed node */
    sn->type = FRONTIER;
    sn->label = dict_insert(d, buf);
  }

  return sn;

}

int elem_label_to_string(elem_label sn, char *s, size_t n) {
  static char buf1[80], buf2[80], buf3[80], buf4[80], buf[80];

  if (sn == NULL) {
    strncpy(s, "NULL", n);
    return 0;
  }

  if (sn->type == FRONTIER && sn->empty) {
    if (show_details) {
      dict_rlookup(d, sn->label, buf1, sizeof(buf1));
      sprintf(s, "%s#", buf1);
    }
  } else if (sn->type == FRONTIER) {
    dict_rlookup(d, sn->label, buf1, sizeof(buf1));
    sprintf(s, "%s", buf1);
  } else {
    dict_rlookup(d, sn->label, buf1, sizeof(buf1));
    if (sn->type == FOOT)
      sprintf(s, "%s*", buf1);
    else if (sn->type == SUBST)
      sprintf(s, "%s+", buf1);
    else
      sprintf(s, "%s", buf1);
  }

  if (show_details) {
    if (sn->noadj)
      strcat(s, "[NA]");

    dict_rlookup(d, sn->headword, buf1, sizeof(buf1));
    dict_rlookup(d, sn->headtag, buf2, sizeof(buf2));
    dict_rlookup(d, sn->d_addr, buf3, sizeof(buf3));
    dict_rlookup(d, sn->pge->d_index, buf4, sizeof(buf4));
    sprintf(buf, " (%s-%s-%s-%s)", buf4, buf2, buf1, buf3);
    strcat(s, buf);
  }

  return 0;
}



