#include "node.h"
#include "elem_label.h"
#include "elem_node.h"

int node_type(node n) {
  elem_label sn = n->contents;
  return sn->type;
}

int node_visible(node n) {
  elem_label sn = n->contents;
  return !sn->virtual;
}

int node_virtual(node n) {
  elem_label sn = n->contents;
  return sn->virtual;
}

int node_empty(node n) {
  elem_label sn = n->contents;
  return sn->empty;
}

int node_label(node n) {
  elem_label sn = n->contents;
  return sn->label;
}

node node_left (node n) {
  return node_sister(n, -1);
}

node node_right (node n) {
  return node_sister(n, 1);
}

int node_adjoinable(node n, node s) {
  elem_label sn = n->contents;
  elem_label ss = s->contents;
  int ntype = sn->pge->treeType, stype = ss->pge->treeType;

  if (ss->noadj)
    return 0;

  if (ss->type == FOOT || ss->type == SUBST)
    return 0;

  if (ss->virtual == LEFT)
    return ntype == LEFTMOD;

  if (ss->virtual == RIGHT)
    return ntype == RIGHTMOD;

  if (ntype == INITIAL)
    return 0;

  if (sn->label != ss->label)
    return 0;

  if (ss->spine && 
      (stype == LEFTMOD || stype == LEFTPRED) &&
      (ntype == RIGHTMOD || ntype == RIGHTPRED))
    return 0;
    
  if (ss->spine && 
      (stype == RIGHTMOD || stype == RIGHTPRED) &&
      (ntype == LEFTMOD || ntype == LEFTPRED))
    return 0;

  return 1;
}

int node_substitutable(node n, node s) {
  elem_label sn = n->contents;
  elem_label ss = s->contents;
  if (sn->label == ss->label && node_parent(n) == NULL && sn->pge->treeType == INITIAL)
    return 1;
  else
    return 0;
}

node find_anchor(node root) {
  node retval;
  int i, n;

  if (root == NULL)
    return NULL;
  else if (((elem_label)root->contents)->anchor)
    return root;
  else {
    n = node_n_children(root);
    for (i=0; i<n; i++) {
      if ((retval = find_anchor(node_child(root, i))) != NULL)
	return retval;
    }
    return NULL;
  }
}

int tree_type (node r) {
  elem_label sr = r->contents;

  return sr->pge->treeType;
}

int node_spine(node n) {
  elem_label sn = n->contents;
  
  return sn->spine;
}

void node_print(node n) {
  char buf[100];
  elem_label_to_string(n->contents, buf, sizeof(buf));
  printf("%s", buf);
}
