/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "grammar.h"
#include "prob.h"
#include "parser.h"

#include "list.h"

#include "node.h"
#include "elem_label.h"
#include "elem_node.h"

/* This is the grammar as read from grammar.txt. It is sentence-independent. */

list_node grammar, subgrammar;

extern int tree_threshold;

extern int grammar_dump;
float besttreeprior[MAX_WORDS];

extern float beam(int span);
float real_p_prior(int d_index, int headtag, int headword);
int hack(node cur);

/*** Functions for grammar_entry ***/

struct grammar_entry * new_grammar_entry(void) {
  struct grammar_entry *pge = malloc(sizeof(*pge));
  pge->root = NULL;
  return pge;
}

void * grammar_entry_delete(struct grammar_entry *pge) {
  tree_delete(pge->root, (DELETE)elem_label_delete);
  free(pge);
  return NULL;
}

struct grammar_entry * grammar_entry_copy(struct grammar_entry *pge) {
  struct grammar_entry *pge2 = malloc(sizeof(*pge2));
  pge2->root = pge->root;
  pge2->treeType = pge->treeType;
  pge2->d_index = pge->d_index;

  return pge2;
}

node tree_root(void *tree) {
  struct grammar_entry *pge = tree;
  return pge->root;
}

int preterminal(node cur) {
  return node_n_children(cur) == 1 &&
    node_type(node_child(cur, 0)) == FRONTIER;
}

/* Look for nodes with terminal nodes that are in w */
int findTerminals(node cur, struct word *w[], int n) {
  int i, flag, nc;
  elem_label sn = node_contents(cur);

  if (sn == NULL || sn->anchor || sn->type == FOOT || sn->type == SUBST)
    return 1;

  if (sn->type == FRONTIER) {
    flag = 0;
    for (i=0; i<n; i++)
      if (w[i]->w == sn->label)
	flag = 1;
    return flag;
  }

  flag = 1;
  nc = node_n_children(cur);
  for (i=0; i<nc; i++)
    flag = flag && findTerminals(node_child(cur, i), w, n);
  return flag;
}

node tree_instantiate(struct grammar_entry *pge, int headword, int oheadword, int headtag) {
  node new = tree_copy(pge->root, (ENDO)elem_label_copy);
  tree_instantiate_helper(pge, new, headword, oheadword, headtag, 1, pge->treeType == LEFTPRED || pge->treeType == RIGHTPRED);

  return new;
}

int tree_instantiate_helper(struct grammar_entry *pge, node cur, int headword, int oheadword, int headtag, int addr, int spine) {
  char buf[80];
  elem_label sn = node_contents(cur), leaf;
  int i, n, child_spine;


  if (sn == NULL)
    return NULL;

  n = node_n_children(cur);
  if (n>2)
    fprintf(stderr, "tree_instantiate(): non-binary branching\n");

  for (i=0; i<n; i++) {
    child_spine = 0;
    if (spine && i == 0 && (pge->treeType == RIGHTMOD || 
			    pge->treeType == RIGHTPRED))
      child_spine = 1;
    if (spine && i == n-1 && (pge->treeType == LEFTMOD || 
			      pge->treeType == LEFTPRED))
      child_spine = 1;
    tree_instantiate_helper(pge, node_child(cur, i), headword, oheadword, headtag, addr*2+i, child_spine);
  }

  sn->pge = pge;
  sn->headword = headword;
  sn->oheadword = oheadword;
  sn->headtag = headtag;

  sprintf(buf, "%d", addr);
  sn->d_addr = dict_lookup(d, buf);

  if (sn->virtual)
    sn->d_type = dict_lookup(d, "MOD");
  else if (sn->type == SUBST) {
    strcpy(buf, "INITIAL/");
    dict_rlookup(d, sn->label, buf+strlen(buf), sizeof(buf)-strlen(buf));
    sn->d_type = dict_lookup(d, buf);
  } else {
    strcpy(buf, "PRED/");
    dict_rlookup(d, sn->label, buf+strlen(buf), sizeof(buf)-strlen(buf));
    sn->d_type = dict_lookup(d, buf);
  }

  sn->spine = spine;

  /* NA at modifier roots, and at subst and foot nodes */
  sn->noadj = /*((pge->treeType == LEFTMOD || pge->treeType == RIGHTMOD) && sn->parent == NULL) || */
    sn->type == SUBST || sn->type == FOOT || sn->type == FRONTIER || sn->anchor || preterminal(cur);

  sn->pp = real_p_prior(pge->d_index, headtag, headword);

  /* Relexicalize if necessary */
  if (sn->anchor && sn->type == INTERIOR) {
    sn->label = headtag;
    if (node_n_children(cur) == 0) {
      node_insert_child(cur, 0, new_node(elem_label_copy(sn)));
      leaf = node_child(cur, 0)->contents;
      leaf->label = oheadword;
      leaf->type = FRONTIER;
      /* Other fields are not right, but it doesn't matter */
    }
  }

  return 0;
}

void tree_register(struct grammar_entry *pge, struct word *w, int j) {
  struct grammar_entry *pge2;
  list_node cur;

  /* First check to see if already in subgrammar */
  for (cur = subgrammar; cur != NULL; cur = cur->next) {
    pge2 = (struct grammar_entry *)cur->contents;
    if (pge2->d_index == pge->d_index &&
	pge2->treeType == pge->treeType &&
	((elem_label)pge2->root->contents)->oheadword == w->ow &&
	((elem_label)pge2->root->contents)->headtag == w->t[j])
      return;
  }
  
  pge2 = grammar_entry_copy(pge);

  pge2->root = tree_instantiate(pge2, w->w, w->ow, w->t[j]);
  hack(pge2->root);
  
  if (grammar_dump)
    print_grammar_entry(pge2);
  
  subgrammar = new_list_node(pge2, subgrammar);
}

void growTree(struct word *w[], int n, int headpos) {
  int j;
  float pp;
  int headword = w[headpos]->w;
  list_node cur;
  struct grammar_entry *pge;
      
  /* Make two passes: find the best tree first, and then actually
     register everything in the beam. */

  besttreeprior[headpos] = IMPOSSIBLE;

  for (cur = grammar; cur != NULL; cur = cur->next) {
    pge = (struct grammar_entry *)(cur->contents);
    if (findTerminals(pge->root, w, n))
      for (j=0; j<w[headpos]->n_t; j++) {
	pp = real_p_prior(pge->d_index, w[headpos]->t[j], headword);
	  
	if (pp > besttreeprior[headpos])
	  besttreeprior[headpos] = pp;
      }
  }
  

  for (cur = grammar; cur != NULL; cur = cur->next) {
    pge = (struct grammar_entry *)(cur->contents);
    if (findTerminals(pge->root, w, n))
      for (j=0; j<w[headpos]->n_t; j++) {
	  pp = real_p_prior(pge->d_index, w[headpos]->t[j], headword);

	  if (pp > besttreeprior[headpos] + beam(1)) {

	    if (0 && pge->treeType == INITIAL) {

	      pge->treeType = RIGHTMOD;

	      tree_register(pge, w[headpos], j);

	      pge->treeType = LEFTMOD; 

	      tree_register(pge, w[headpos], j);

	      pge->treeType = INITIAL;
	    } 

	    tree_register(pge, w[headpos], j);

	  }
      }
  }
}

list_node growTrees(struct word *w[], int n) {
  int i;

  subgrammar = NULL;

  for (i=0; i<n; i++)
    growTree(w, n, i);

  return subgrammar;
}

void deleteTrees(list_node subgrammar) {
  list_map(subgrammar, (ENDO)grammar_entry_delete);
  list_delete(subgrammar);
}

int determine_type(node root) {
  elem_label sr = node_contents(root);
  int left, right;

  if (root == NULL)
    return INITIAL;

  if (sr->type == FOOT)
    return -1;
  left = determine_type(node_child(root, 0));
  right = determine_type(node_child(root, 1));
  if ((left == RIGHTPRED || left == -1) && right == INITIAL)
    return RIGHTPRED;
  if ((left == LEFTPRED || left == -1) && node_n_children(root) == 1)
    return LEFTPRED;
  if ((right == LEFTPRED || right == -1) && left == INITIAL)
    return LEFTPRED;
  if (left == INITIAL && right == INITIAL)
    return INITIAL;
  return -2;
}

struct grammar_entry * read_grammar_entry(FILE *fp) {
  int index, count;
  static char buf[1000];
  node root;
  struct grammar_entry *result;

  if (fscanf(fp, "%d %d ", &index, &count) < 2)
    return NULL;

  if (count < tree_threshold) {
    fgets(buf, sizeof(buf), fp);
    return read_grammar_entry(fp);
  }

  root = read_tree(fp, (FROM_STRING)elem_nonterm_from_string, (FROM_STRING)elem_term_from_string);

  if (root == NULL)
    return NULL;

  result = new_grammar_entry();

  result->root = root;

  result->treeType = determine_type(root);

  sprintf(buf, "%d", index);
  result->d_index = dict_lookup(d, buf);

  return result;
}

int print_grammar_entry(struct grammar_entry *pge) {
  char buf[80];
  switch(pge->treeType) {
  case INITIAL: printf ("Initial tree"); break;
  case LEFTPRED: printf ("Left predicative tree"); break;
  case RIGHTPRED: printf ("Right predicative tree"); break;
/*  case LEFTMOD: printf ("Left modifier tree"); break;
  case RIGHTMOD: printf ("Right modifier tree"); break;*/
  }
  dict_rlookup(d, pge->d_index, buf, sizeof(buf));
  printf (" (index = %s)\n", buf);

  tree_print(stdout, pge->root, (TO_STRING)elem_label_to_string);
  printf("\n");
}

int read_grammar(const char *prefix) {
  FILE *fp;
  struct grammar_entry *result;
  char filename[FILENAME_MAX];

  sprintf(filename, "%s.grm", prefix);
  fp = fopen(filename, "r");

  grammar = NULL;
  result = read_grammar_entry(fp);
  while (result != NULL) {
    grammar = new_list_node(result, grammar);
    result = read_grammar_entry(fp);
  }

  fclose(fp);

  return 0;
}

void delete_grammar(void) {
  list_map(grammar, (ENDO)grammar_entry_delete);
  list_delete(grammar);
}
