/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>

#include "dict.h"
#include "prob.h"
#include "profile.h"

#include "node.h"
#include "elem_label.h"
#include "elem_node.h"

#include "grammar.h"
#include "model.h"

/* Global options */

char prefix[FILENAME_MAX] = "tig0", derivation_filename[FILENAME_MAX];
int derivation_dump = 0, left_to_right = 1, code_level = 2;

prob probs[2], p_p1;

code_t CODE_START, CODE_STOP, CODE_MID;
int stop, nil;
int top, none, root;

#define N_KEYS 11

FILE *deriv_fp;

dict d;

void itoa (char *buf, int n) {
  static char temp[20];
  int i = 0, j;

  while (n>0) {
    temp[i] = '0'+n%10;
    n = n/10;
    i++;
  }

  if (i==0) {
    temp[i] = '0';
    i++;
  }

  for (j=0; j<i; j++)
    buf[i-j-1] = temp[j];
  buf[i] = '\0';

}

float p_initial(node r) {
  elem_label sr = node_contents(r);
  static int ks[N_KEYS];
  prob cur;
  float p;

  ks[0] = sr->pge->d_index;
  ks[1] = sr->headtag;
  ks[2] = sr->headword;
  ks[3] = ks[4] = ks[5] = ks[6] = ks[7] = ks[8] = ks[9] = ks[10] = nil;

  p = 0.0;
  for (cur = probs[0]; cur != NULL; cur = cur->next)
    p += prob_mapped_lookup(cur, ks);
  return p;
}

code_t compute_code(code_t code, node n) {
  if (left_to_right) {
    code.code = -1;
    code.node = n;
  } else if (code_level == 2) {
    code.code = ((elem_label)n->contents)->label;
  } else if (code_level == 1) {
    code.code = CODE_MID.code;
  } else {
    code.code = CODE_START.code;
  }

  return code;
}

int code_equal(code_t code1, code_t code2) {
  if (code1.code == -1 && code2.code == -1 && left_to_right)
    return code1.node == code2.node;
  else
    return code1.code == code2.code;
}

void code_string(code_t code, char *s, int n) {
  if (code.code == -1 && left_to_right)
    dict_rlookup(d, ((elem_label)code.node->contents)->label, s, n);
  else
    dict_rlookup(d, code.code, s, n);
}

float real_p_adj(node r, node s, int c) {
  elem_label sr = node_contents(r), ss = node_contents(s);
  node parent;
  static int ks[N_KEYS];
  prob cur;
  float p;

  parent = s;
  while (node_virtual(parent))
    parent = node_parent(parent);

  if (sr == NULL) {
    /* no adjunction has probability 1 for NA nodes */
    if (ss->noadj || ss->type == FRONTIER) /* might not be valid for
					      other tweaks */
      return 0.0;

    ks[0] = stop;
    ks[1] = stop;
    ks[2] = stop;
    ks[3] = ss->pge->d_index;
    ks[4] = ss->headtag;
    ks[5] = ss->headword;
    ks[6] = ss->d_type;
    ks[7] = ss->d_addr;
    ks[8] = ((elem_label)parent->contents)->label;
    ks[9] = c;
    ks[10] = nil;

    p = 0.0;
    for (cur = probs[0]; cur != NULL; cur = cur->next)
      p += prob_mapped_lookup(cur, ks);
    return p;
  }

/*  if (sr->pge->treeType == INITIAL)
    return IMPOSSIBLE;*/

  ks[0] = sr->pge->d_index;
  ks[1] = sr->headtag;
  ks[2] = sr->headword;
  ks[3] = ss->pge->d_index;
  ks[4] = ss->headtag;
  ks[5] = ss->headword;
  ks[6] = ss->d_type;
  ks[7] = ss->d_addr;
  ks[8] = ((elem_label)parent->contents)->label;
  ks[9] = c;
  ks[10] = nil;
  
  p = 0.0;
  for (cur = probs[0]; cur != NULL; cur = cur->next)
    p += prob_mapped_lookup(cur, ks);
  return p;
}

float p_adj(node r, node s, code_t code) {
  elem_label sr = node_contents(r), ss = node_contents(s);
  int c;

  if (code.code == -1 && left_to_right) {
    if (code_level == 2)
      c = ((elem_label)code.node->contents)->label;
    else if (code_level == 1)
      c = CODE_MID.code;
    else
      c = CODE_START.code;
  } else
    c = code.code;

  if (ss->virtual == LEFT && left_to_right) {
    if (sr == NULL)
      c = CODE_START.code;
    else {
      if (code_level == 2)
	c = sr->label;
      else if (code_level == 1)
	c = CODE_MID.code;
      else
	c = CODE_START.code;
    }

    if (code.code == -1)
      r = code.node;
    else
      r = NULL;
  }

  return real_p_adj(r, s, c);
}

float p_subst(node r, node s) {
  elem_label sr = node_contents(r), ss = node_contents(s);
  static int ks[N_KEYS];
  float p;
  prob cur;

  if (sr->pge->treeType != INITIAL)
    return IMPOSSIBLE;

  ks[0] = sr->pge->d_index;
  ks[1] = sr->headtag;
  ks[2] = sr->headword;
  ks[3] = ss->pge->d_index;
  ks[4] = ss->headtag;
  ks[5] = ss->headword;
  ks[6] = ss->d_type;
  ks[7] = ss->d_addr;
  ks[8] = ss->label;
  ks[9] = CODE_START.code;
  ks[10] = nil;

  p = 0.0;
  for (cur = probs[0]; cur != NULL; cur = cur->next)
    p += prob_mapped_lookup(cur, ks);
  return p;

}

float p_prior(node n, code_t code) {
  elem_label sn = node_contents(n);

  if (sn->virtual == LEFT && code.code == -1 && left_to_right)
    return sn->pp + real_p_adj(code.node, n, CODE_START.code);

  return sn->pp;
}

float real_p_prior(int d_index, int headtag, int headword) {
  static int ks[N_KEYS];
  prob cur;
  float p;

  ks[0] = d_index;
  ks[1] = headtag;
  ks[2] = headword;
  ks[10] = nil;
  
  p = 0.0;
  for (cur = probs[1]; cur != NULL; cur = cur->next)
    p += prob_mapped_lookup(cur, ks);
  return p;
  
}

int hack(node cur) {
  int dom_anchor;
  int ks[N_KEYS];
  int i, n;
  elem_label sn;

  if (cur == NULL)
    return 0;
  
  sn = cur->contents;

  if (sn->anchor)
    return 1;

  n = node_n_children(cur);
  dom_anchor = 0;
  for (i=0; i<n; i++)
    dom_anchor = dom_anchor || hack (node_child(cur, i));

  ks[0] = sn->headtag;
  ks[1] = sn->headword;
  ks[2] = nil;

  if (!dom_anchor)
    sn->pp -= prob_lookup(p_p1, ks);

  return dom_anchor;
}

int model_init(void) {
  FILE *dictfp;
  int tables, i, table;
  prob cur;
  char name[100];
  int c;

  static char *names[N_KEYS] = {
    "tree'", "t'", "w'",
    "tree", "t", "w",
    "type", "addr", "label", "code",
    "NIL" };


#define HUGE_CACHE 400000
#define BIG_CACHE 80000
#define LITTLE_CACHE 20000

  tables = read_profile(prefix, probs, 2);

  table = 0;
  for (i=0; i<tables; i++) {
    for (cur = probs[i]; cur != NULL; cur = cur->next) {
      sprintf(name, "%s.%d", prefix, table);
      prob_read(cur, name);
      table++;
    }
  }

  for (cur = probs[0]; cur != NULL; cur = cur->next) {
    prob_set_cache(cur, HUGE_CACHE);
    prob_set_mapping(cur, names, N_KEYS);
  }

  p_p1 = probs[1];  /* for hack(), ugh */

  for (cur = probs[1]; cur != NULL; cur = cur->next) {
    prob_set_cache(cur, LITTLE_CACHE);
    prob_set_mapping(cur, names, N_KEYS);
  }

  sprintf(name, "%s.dict", prefix);
  dictfp = fopen (name, "r");
  d = read_dict(dictfp); 
  fclose(dictfp);

  dict_insert (d, "@");
  top = dict_lookup(d, "TOP");
  none = dict_lookup(d, "-NONE-");
  CODE_START.code = dict_insert(d, "*START*");
  CODE_MID.code = dict_lookup(d, "*MID*");
  CODE_STOP.code = stop = dict_lookup(d, "*STOP*");
  nil = dict_lookup(d, "NIL");
  root = dict_lookup(d, "1");

  read_grammar(prefix);

  if (derivation_dump)
    deriv_fp = fopen(derivation_filename, "w");
  else
    deriv_fp = NULL;

  return 0;

}

int done(void) {

  int i;
  prob cur, next;

  for (i=0; i<2; i++)
    for (cur=probs[i]; cur!=NULL; cur=next) {
      next = cur->next;
      prob_delete(cur);
    }

  delete_grammar();
  dict_delete(d);
  
  return 0;
}



void node_stats(void) {
}


extern char *words[];

void print_attachment(node n, node s) {
  elem_label sn = node_contents(n), ss = node_contents(s);
  static char buf[80];

  if (sn) {
    dict_rlookup(d, sn->headtag, buf, sizeof(buf));
/*    fprintf(deriv_fp, "%d %d %s ", sn->headpos+1, sn->index, buf);*/
    if (ss) {
      dict_rlookup(d, ss->headtag, buf, sizeof(buf));
/*      fprintf (deriv_fp, "%d %d %d %s\n", ss->headpos+1, ss->addr, ss->index, buf);*/
    }
    else {
      fprintf (deriv_fp, "NIL NIL NIL NIL\n");
    }
  } else {
    fprintf (deriv_fp, "0 0 0 0 0 0 0\n");
  }
}

void print_eod(void) {
  fprintf(deriv_fp, ".\n");
}
