#include <stdlib.h>

#include "generic.h"
#include "hash_map.h"
#include "chart.h"

#include "item.h"

extern int trace;

chart new_chart(void) {

  chart c;
  c = malloc(sizeof (*c));
  c->chart_size = c->total_size = 0;
  c->max_size = 1000;
  c->items = malloc(c->max_size * sizeof(item));

  c->r_index = new_hash_map((COMPARE)item_compare, (HASH)item_r_hash);
  c->a_index = new_hash_map((COMPARE)item_compare, (HASH)item_a_hash);

  c->goal = NULL;

  return c;
}

void chart_delete(chart c) {
  int i;
  hash_map_delete(c->r_index);
  hash_map_delete(c->a_index);
  for (i=0; i<c->total_size; i++)
    item_delete(c->items[i]);
  item_delete(c->goal);
  free(c->items);
  free(c);
}

int chart_add_goal(chart c, item it) {
  if (trace) {
    printf("Adding goal ");
  }

  if (c->goal == NULL) {
    c->goal = it;
    if (trace)
      printf("\n");
  } else {
    item_merge(c->goal, it);
    if (trace)
      printf("(already added)\n");
  }

  return 0;
}

int chart_add(chart c, item it) {
  item *itp;

  if (trace) {
    printf("Adding      ");
    item_print(stdout, it);
  }

  itp = (item *)hash_map_lookup(c->r_index, it, 1);
  if (*itp == NULL) {

    if (c->total_size >= c->max_size) {
      c->max_size *= 2;
      c->items = realloc(c->items, c->max_size * sizeof(item));
    }
    c->items[c->total_size] = it;

    if (trace)
      printf(" (%d)\n", c->total_size);

    c->total_size++;

    *itp = it;
  } else {
    item_merge(*itp, it);
    if (trace)
      printf(" (already in chart)\n");
  }

  return 0;
}

int chart_parse (chart c) {
  item trigger, *itp;

  while (c->chart_size < c->total_size) {
    trigger = c->items[c->chart_size];
    if (trace) {
      printf("Considering ");
      item_print(stdout, trigger);
      printf(" (%d)\n", c->chart_size);
    }
    c->chart_size++; /* move trigger to chart */
    itp = (item *)hash_map_lookup(c->a_index, trigger, 1);
    if (*itp == NULL)
      *itp = trigger;
    else
      fprintf(stderr, "duplicate items in chart and agenda\n");
    add_consequences(c, trigger);
  }
}
