/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdlib.h>
#include <stdio.h>

#include "node.h"
#include "tag.h"
#include "lexicon.h"

#define CODE_START "*START*"
#define CODE_MID "*MID*"

extern int delexicalize, binarize, left_to_right, code_level;

int assign_addresses(node cur, int addr) {
  int i, n;
  int h = -1;

  ((derived_label)cur->contents)->addr = addr;
  n = node_n_children(cur);

  if (!binarize) {
    addr++;
    for (i=0; i<n; i++)
      addr = assign_addresses(node_child(cur, i), addr);
  } else {
    /* Skip past adjunction sites which are going to get removed later */
    if (n == 1 && ((derived_label)cur->contents)->type >= 0) {
      return assign_addresses(node_child(cur, 0), addr);
    }

    for (i=0; i<n; i++)
      if (((derived_label)node_child(cur, i)->contents)->head ||
	  /* the head flag lives in the bottom half of the adjunction site */
	  node_n_children(node_child(cur, i)) == 1 &&
	  ((derived_label)node_child(cur, i)->contents)->type >= 0 &&
	  ((derived_label)node_child(node_child(cur, i), 0)->contents)->head) {
	h = i;
	break;
      }
    
    if (n > 0) {
      addr *= 2;
      for (i=0; i<h; i++) {
	assign_addresses(node_child(cur, i), addr*2);
	addr = addr * 2 + 1;
      }
      addr *= 2;
      for (i=n-1; i>h; i--) {
	assign_addresses(node_child(cur, i), addr*2+1);
	addr = addr * 2;
      }
      addr *= 2;
      assign_addresses(node_child(cur, h), addr);
    }
  }

  return addr;
} 

void print_tree_name(derivation_label dl, int primed, lexicon words, int threshold) {
  if (dl == NULL) {
    if (primed)
      printf("tree'=*STOP* t'=*STOP* w'=*STOP* ");
    else
      printf("tree=NIL t=NIL w=NIL ");
  } else {
    if (primed)
      printf("tree'=%d t'=%s w'=%s ", dl->elementary_tree->index, dl->headtag, fold_unknown(dl->headword, words, threshold));
    else
      printf("tree=%d t=%s w=%s ", dl->elementary_tree->index, dl->headtag, fold_unknown(dl->headword, words, threshold));
  }
}

void print_addr(node presite, node postsite) {
  int addr, slot;
  node site;

  if (presite == postsite) {
    addr = ((derived_label)presite->contents)->addr;

    if (node_n_children(presite) != 0)
      printf("type=PRED/%s label=%s ", ((derived_label)presite->contents)->label, ((derived_label)presite->contents)->label);
    else
      printf("type=INITIAL/%s label=%s ", ((derived_label)presite->contents)->label, ((derived_label)presite->contents)->label);

    printf("addr=%d ", addr);
    
  } else if (!binarize) {
    if (node_order(postsite) == 0) {
      site = node_parent(postsite);
      addr = ((derived_label)site->contents)->addr;
      slot = 0;
    } else {
      site = node_parent(presite);
      addr = ((derived_label)site->contents)->addr;
      slot = node_order(presite)+1;
    }

    printf("type=MOD label=%s ", ((derived_label)site->contents)->label);
    printf("addr=%d.%d ", addr, slot);
  } else {
    if (node_parent(presite) == node_parent(postsite)) {
      site = node_parent(presite);
      if (((derived_label)postsite->contents)->addr % 2 != 0)
	addr = ((derived_label)postsite->contents)->addr - 1;
      else 
	addr = ((derived_label)presite->contents)->addr + 1;
    } else if (node_order(postsite) == 0) {
      /* Leftmost slot */
      site = node_parent(postsite);
      addr = ((derived_label)node_parent(postsite)->contents)->addr * 2;
    } else {
      /* Rightmost slot */
      site = node_parent(presite);
      addr = ((derived_label)presite->contents)->addr / 2;
    }

    printf("type=MOD label=%s ", ((derived_label)site->contents)->label);
    printf("addr=%d ", addr);

  }
  
}

const char * compute_code(const char *code, node mod) {
  node root = ((derivation_label)mod->contents)->elementary_tree->root;
  static char buf[100];

  if (code_level == 0)
    return CODE_START;
  else if (code_level == 1)
    return CODE_MID;
  else if (code_level == 2) {
    if (((derived_label)root->contents)->anchor)
      return ((derivation_label)mod->contents)->headtag;
    else {
      return ((derived_label)root->contents)->label;
    }
  }
}

void count_mods(node deriv, node cur, lexicon words, int threshold) {
  int i, n, j, m, h=-1;
  const char *code;
  node child;

  n = node_n_children(deriv);
  m = node_n_children(cur);

  /* Emit modifiers for each sister-adjunction slot. Note that this
     code relies on derivation-precedence corresponding to
     derived-precedence. */

  if (!((derived_label)cur->contents)->anchor && m > 0 &&
      ((derived_label)cur->contents)->type == TYPE_OTHER) {
    for (j=0; j<m; j++)
      if (((derived_label)node_child(cur, j)->contents)->head) {
	h = j;
	break;
      }

    for (j=0; j<m+1; j++) {
      code = CODE_START;
      
      for (i=0; i<n; i++) {
	if (j<=h && !left_to_right)
	  child = node_child(deriv, n-i-1);
	else
	  child = node_child(deriv, i);

	if (((derivation_label)child->contents)->optional &&
	    ((derivation_label)child->contents)->site == cur &&
	    ((derivation_label)child->contents)->slot == j) {
	  count_events(child, code, words, threshold);
	  code = compute_code(code, child);
	}
      }

      /* Emit stop event */

      printf("1.0 ");
      print_tree_name(NULL, 1, words, threshold);
      print_addr(j==0?NULL:node_child(cur, j-1), node_child(cur, j));
      printf("code=%s ", code);
      print_tree_name(deriv->contents, 0, words, threshold);
      
      printf("NIL=NIL\n");
    }
  }

  for (j=0; j<m; j++)
    count_mods(deriv, node_child(cur, j), words, threshold);

}


int count_events(node deriv, const char *code, lexicon words, int threshold) {
  node site1 = NULL, site2 = NULL; 
  derivation_label dl = ((derivation_label)deriv->contents), pdl;
  int i, n;

  if (node_parent(deriv) != NULL) {
    pdl = ((derivation_label)node_parent(deriv)->contents);

    if (!dl->optional) {
      site1 = site2 = find_site(pdl->elementary_tree->root, node_order(deriv));
    } else {
      site1 = dl->slot>0 ? node_child(dl->site, dl->slot-1) : NULL;
      site2 = dl->slot<node_n_children(dl->site) ? node_child(dl->site, dl->slot) : NULL;
    }
  } else {
    pdl = NULL;
  }

  /* Emit modification event */

  printf("1.0 ");
  if (tree_trivial(dl->elementary_tree))
    print_tree_name(NULL, 1, words, threshold);
  else
    print_tree_name(dl, 1, words, threshold);
  print_tree_name(pdl, 0, words, threshold);
  if (pdl != NULL) {
    print_addr(site1, site2);
    printf("code=%s ", code);
  } else {
    printf("addr=NIL type=NIL label=NIL code=NIL ");
  }
  printf("NIL=NIL\n");
  
  assign_addresses(((derivation_label)deriv->contents)->elementary_tree->root, 1);

  count_mods(deriv, ((derivation_label)deriv->contents)->elementary_tree->root, words, threshold);

  /* Finally, get all the non-optional modifiers */
  n = node_n_children(deriv);
  for (i=0; i<n; i++)
    if (!((derivation_label)node_child(deriv, i)->contents)->optional)
      count_events(node_child(deriv, i), CODE_START, words, threshold);

  return 0;
}

