/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdlib.h>
#include <string.h>

#include "node.h"
#include "tag.h"

node extract_single(node cur, node other, node deriv, int state) {
  int i, n, new_state;
  node new;

/* In state 0, cur is the remnant and other is the extract;
   in state 1, other is the remnant and cur is the extract. */

  new_state = (((derived_label)cur->contents)->elementary_tree == ((derivation_label)deriv->contents)->elementary_tree);

  if (state != new_state) {
    /* Move the subtree rooted at cur over to other, leaving behind
       an adjunction node or a foot node, as appropriate. */

    new = new_node(derived_label_copy(cur->contents));

    node_insert_child(node_parent(cur), node_order(cur), new);
    
    if (state == 0) {
      if (node_parent(deriv) != NULL) {
	((derived_label)new->contents)->type = node_order(deriv);
	((derived_label)new->contents)->elementary_tree = ((derivation_label)node_parent(deriv)->contents)->elementary_tree;
	((derived_label)new->contents)->trivials = 0;
      }
    } else {
      ((derived_label)new->contents)->type = -1; /* foot */
      ((derived_label)new->contents)->elementary_tree = ((derivation_label)deriv->contents)->elementary_tree;
      ((derived_label)new->contents)->trivials = 0;
    }
    
    node_insert_child(other, -1, cur);
    other = new;

    /* cur and other are swapped now */
  } else {
    new = cur;
  }

  n = node_n_children(cur);
  for (i=0; i<n; i++)
    extract_single(node_child(cur, i), other, deriv, new_state);

  return new;
}

int trim_single(node cur, node deriv) {
  derivation_label dl;
  int i;

  i = 0;
  while (i < node_n_children(cur))
    i += trim_single(node_child(cur, i), deriv);

  if (cur->contents == NULL || ((derived_label)cur->contents)->type == -2) /* dummy or non-adjunction node */
    return 1;

  dl = (derivation_label)node_child(deriv, ((derived_label)cur->contents)->type)->contents;

  if (node_n_children(cur) == 0 && dl->optional) {
    dl->site = node_parent(cur);
    dl->slot = node_order(cur);
    node_delete(cur);
    return 0;
  }
  return 1;
}

int tree_trivial(elementary_tree et) {
  return ((derived_label)et->root->contents)->type == TYPE_FOOT;
}

node extract_trivials(node cur, node deriv) {
  int i, n;
  int order;
  node new, retval;
  derived_label ddl;
  derivation_label dnl;
  elementary_tree et;

  n = node_n_children(cur);

  retval = cur;

  if (((derived_label)cur->contents)->elementary_tree == ((derivation_label)deriv->contents)->elementary_tree)
    for (i=0; i<((derived_label)cur->contents)->trivials; i++) {
      et = new_elementary_tree();
      
      dnl = new_derivation_label();
      dnl->elementary_tree = et;
      new = new_node(dnl);
      node_insert_child(deriv, -1, new);
      order = node_order(new);
      
      ddl = derived_label_copy(cur->contents);
      ddl->elementary_tree = et;
      ddl->type = TYPE_FOOT;
      et->root = new_node(ddl);

      ddl = derived_label_copy(cur->contents);
      ddl->type = order;
      new = new_node(ddl);
      node_supplant(cur, new);
      node_insert_child(new, 0, cur);

      if (retval == cur)
	retval = new;
    }

  for (i=0; i<n; i++)
    extract_trivials(node_child(cur, i), deriv);
  return retval;
}

node extract(node cur, node deriv) {
  int i, n;
  node new;

  n = node_n_children(deriv);
  for (i=0; i<n; i++)
    cur = extract(cur, node_child(deriv, i));

  cur = extract_trivials(cur, deriv);

  new = new_node(NULL); /* dummy node */
  cur = extract_single(cur, new, deriv, 0);
  trim_single(new, deriv);

  if (node_n_children(new) == 0) {
    fprintf(stderr, "Warning: empty tree extracted\n");
    ((derivation_label)deriv->contents)->elementary_tree->root = NULL;
  } else if (node_n_children(new) > 1) {
    fprintf(stderr, "Warning: multicomponent set not rooted\n");
  } else {
    ((derivation_label)deriv->contents)->elementary_tree->root = node_child(new, 0);
  }

  node_delete(new);

  /* If deriv is root, then cur is just an empty shell -- probably
     just a single node */
  if (node_parent(deriv) == NULL) {
    tree_delete(cur, (DELETE)derived_label_delete);
    cur = NULL;
  }

  return cur;
}

