#include <stdlib.h>
#include <stdio.h>

#include <string.h>

#include "lexicon.h"
#include "generic.h"
#include "hash_map.h"

#define whitespace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n')

static int next_id = 0, max_id;
static void **objects = NULL;

lexicon new_lexicon(_COMPARE(compare), _HASH(hash)) {
  if (objects == NULL) {
    max_id = 10;
    objects = malloc(max_id * sizeof(void *));
  }

  return new_hash_map(compare, hash);
}

int lexicon_size(lexicon lex) {
  return lex->nelem;
}

void * lexicon_lookup(lexicon lex, void *x) {
  struct lex_entry **pple;
  pple = (struct lex_entry **)hash_map_lookup(lex, x, 0);
  if (pple == NULL)
    return NULL;
  else
    return (*pple)->key;
}

/* If x is not already in lex, returns x.
   If x is already in lex, returns the one that is in lex. */

void * lexicon_insert(lexicon lex, void *x) {
  struct lex_entry **pple;

  pple = (struct lex_entry **)hash_map_lookup(lex, x, 1);

  if (*pple == NULL) {
    *pple = malloc(sizeof(struct lex_entry));

    (*pple)->id = next_id;
    (*pple)->count = 1;
    (*pple)->key = x;

    while (next_id >= max_id) {
      max_id = max_id*2;
      objects = realloc(objects, max_id * sizeof(void *));
    }

    objects[next_id] = x;
    next_id++;

    return x;
  } else {
    (*pple)->count++;
    return (*pple)->key;
  }
}

int lexicon_count(lexicon lex, void *x) {
  struct lex_entry **pple;

  pple = (struct lex_entry **)hash_map_lookup(lex, x, 0);

  if (pple == NULL)
    return 0;
  else
    return (*pple)->count;
}

int lexicon_id(lexicon lex, void *x) {
  struct lex_entry **pple;

  pple = (struct lex_entry **)hash_map_lookup(lex, x, 0);

  if (pple == NULL)
    return -1;
  else
    return (*pple)->id;
}

void * lexicon_retrieve(lexicon lex, int id) {
  return objects[id];
}

int lexicon_write(lexicon lex, FILE *fp, _WRITE(writer)) {
  unsigned int hash;
  list_node cur;
  struct lex_entry *ple;

  for (hash = 0; hash<lex->size; hash++)
    for (cur = lex->bucket[hash]; cur != NULL; cur = cur->next) {
      ple = (struct lex_entry *)((struct hash_map_entry *)cur->contents)->value;
      fprintf(fp, "%d %d ", ple->id, ple->count);
      writer(fp, ((struct hash_map_entry *)cur->contents)->key);
      fprintf(fp, "\n");
    }
}

int lexicon_read(lexicon lex, FILE *fp, _READ(reader)) {
  int id, count;
  struct lex_entry **pple;
  void *x;

  while (!feof(fp)) {
    if (fscanf(fp, "%d%d", &id, &count) < 2)
      break;
    x = reader(fp);
    pple = (struct lex_entry **)hash_map_lookup(lex, x, 1);
    if (*pple != NULL)
      fprintf(stderr, "lexicon_read(): warning: key %d already present\n", id);
    else {
      *pple = malloc(sizeof(struct lex_entry));

      (*pple)->id = id;
      (*pple)->count = count;
      (*pple)->key = x;

      while (id >= max_id) {
	max_id = max_id*2;
	objects = realloc(objects, max_id * sizeof(void *));
      }

      objects[id] = x;
      if (id > next_id)
	next_id = id+1;
    }
  }
  return 0;
}

void string_write(FILE *fp, char *s) {
  fprintf(fp, "%s", s);
}

char * string_read(FILE *fp) {
  char *s;
  int c, max = 10, i;
  s = malloc(max * sizeof(char));

  i = 0;

  c = fgetc(fp);
  while (c != EOF && whitespace(c)) {
    c = getc(fp);
  }

  while (c != EOF && !whitespace(c)) {
    s[i] = c;
    i++;
    if (i>=max) {
      max = max*2;
      s = realloc(s, max*sizeof(char));
    }
    c = fgetc(fp);
  }

  s[i] = '\0';

  return s;
}

int string_hash(char *s) {
  return 0;
}

lexiconIterator new_lexiconIterator(lexicon l) {
  lexiconIterator li = malloc(sizeof (*li));

  li->l = l;
  li->hash = 0;
  li->cur = li->l->bucket[li->hash];
  while (li->cur == NULL && li->hash < li->l->size) {
    li->hash++;
    if (li->hash < li->l->size)
      li->cur = li->l->bucket[li->hash];
    else
      li->cur = NULL;
  }

  return li;
}

int lexiconIterator_hasNext(lexiconIterator li) {
  return li->cur != NULL;
}

void * lexiconIterator_next(lexiconIterator li) {
  void *x;
  x = ((struct hash_map_entry *)li->cur->contents)->key;

  li->cur = li->cur->next;
  while (li->cur == NULL && li->hash < li->l->size) {
    li->hash++;
    if (li->hash < li->l->size)
      li->cur = li->l->bucket[li->hash];
    else
      li->cur = NULL;
  }

  return x;
}

