/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdlib.h>
#include <string.h>

#include "node.h"
#include "tag.h"
#include "treebank_label.h"
#include "tables.h"

extern int delexicalize, binarize, predicatives;
const char anchor[] = "@";

int right_corner(node anc, node desc) {
  if (anc == desc)
    return 1;
  else if (anc == NULL)
    return 0;
  else return right_corner(node_child(anc, -1), desc);
}

int punct(node cur) {
  const char *label = ((treebank_label)cur->contents)->label;
  return (!strcmp(label, ".") || !strcmp(label, ",") || !strcmp(label, ":"));
}

int preterm(node cur) {
  return (node_n_children(cur) == 1 && 
	  node_n_children(node_child(cur, 0)) == 0);
}

void check_coord(node cur, int *ph, int *a) {
  int n, i, c, *code, f, l, r;
  n = node_n_children(cur);

  code = malloc(n * sizeof(int));

  /* Look for conjunction */
  c = -1;
  for (i=0; i<n; i++) {
    if (!strcmp(((treebank_label)node_child(cur, i)->contents)->label, "CC")) {
      c = i;
      code[i] = 1; /* CC */
    } else if (punct(node_child(cur, i)))
      code[i] = 2; /* Punct */
    else if (preterm(node_child(cur, i)))
      code[i] = 3; /* single word modifier */
    else if (!strcmp(((treebank_label)node_child(cur, i)->contents)->label,
		     ((treebank_label)cur->contents)->label))
      code[i] = 4; /* possible conjunct */
    else
      code[i] = 5;
  }

  /* funny cases:
     NP -> DT NP and NP (87x) (either...or, neither...nor, both...and)
     NP -> NP and RB NP (27x) (but not, and even)
     VP -> VP and RB VP (25x) (and then)
     */

  if (c >= 0) {

    /* Don't process flat coordination structures */
/*    for (i=0; i<n; i++) {
      if (code[i] == 3) {
	free(code);
	return;
      }
    }*/

    /* Don't process phrases starting with conjunctions. */

    l = r = -1;
    for (i=c-1; i>=0; i--)
      if (code[i] == 4)
	l = i;
    for (i=c+1; i<n; i++)
      if (code[i] == 4)
	r = i;

    if (l < 0 || r < 0) {
      free(code);
      return;
    }

    *ph = c;

    for (i=0; i<n; i++)
      a[i] = 0;
    a[l] = a[r] = 1;
  }

  free(code);

}

node find_heads(node cur) {
  int i, h=-1, n, *a = NULL;
  derived_label dl;
  treebank_label tl;
  node arg, foot;
  node deriv;

  n = node_n_children(cur);

  if (n > 0) {
    /* Do this first before cur gets altered */

    h = head_table_lookup(cur);
    if (h < 0) {
      fprintf(stderr, "find_heads(): head rules failed\n");
      return NULL;
    }

    a = malloc(n * sizeof(int));
    arg_table_lookup(cur, h, a);

    check_coord(cur, &h, a);

    /* Copy structure */

    for (i=0; i<n; i++)
      find_heads(node_child(cur, i));
  }

  /* Copy Treebank label into new label */
  dl = new_derived_label();
  tl = (treebank_label)node_contents(cur);

  dl->label = strdup(tl->label);

  /* Build up derivation */

  if (n > 0) {
    deriv = node_root(((derived_label)node_child(cur, h)->contents)->percolated);
    ((derived_label)node_child(cur, h)->contents)->head = 1;
    if (delexicalize == 2 && ((derived_label)node_child(cur, h)->contents)->anchor == 1) {
      ((derivation_label)deriv->contents)->headtag = dl->label;
      dl->label = strdup(anchor);
      dl->anchor = 2;
    } else {
      dl->trivials = 1;
    }
  } else {
    deriv = new_node(new_derivation_label());
    ((derivation_label)deriv->contents)->elementary_tree = new_elementary_tree();
    if (delexicalize) {
      ((derivation_label)deriv->contents)->headword = dl->label;
      dl->label = strdup(anchor);
      dl->anchor = 1;
    }
  }


  for (i=0; i<n; i++)
    if (i != h) {
      node_insert_child(deriv, -1, node_root(((derived_label)node_child(cur, i)->contents)->percolated));
      ((derived_label)node_child(cur, i)->contents)->elementary_tree->root = node_child(cur, i); /* for bookkeeping only */
      if (!a[i])
	((derivation_label)node_child(deriv, -1)->contents)->optional = 1;
    }

  /* Finish building label */

  dl->percolated = deriv; /* this may change later */
  dl->elementary_tree = ((derivation_label)deriv->contents)->elementary_tree;

  cur->contents = dl;
  treebank_label_delete(tl);

  /* Try to form a predicative tree */

  n = node_n_children(deriv);
  for (i=0; i<n; i++) {
    arg = node_child(deriv, i);
    foot = ((derivation_label)arg->contents)->elementary_tree->root;
    if (predicatives &&
	!((derivation_label)arg->contents)->optional &&
	right_corner(cur, foot) &&
	!strcmp(((derived_label)foot->contents)->label, "VP") &&
	!strcmp(((derived_label)foot->contents)->label, ((derived_label)cur->contents)->label)) {


      dl->elementary_tree->root = cur;
      ((derived_label)foot->contents)->elementary_tree->root = cur;

      ((derived_label)foot->contents)->head = 1;

      ((derived_label)foot->contents)->trivials = 0;
      
      node_detach(arg);
      node_insert_child(((derived_label)foot->contents)->percolated, -1, deriv);
      break;
    }
  }
  
  free (a);
  return dl->percolated;
}
