/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#include "treebank_label.h"
#include "lexicon.h"
#include "unknown.h"
#include "node.h"

lexicon d;

void process(node cur) {
  int i, j, n, found;
  treebank_label tl1, tl2;
  node other;
  
  n = node_n_children(cur);

  if (n == 1 && node_n_children(node_child(cur, 0)) == 0) {
    other = lexicon_insert(d, cur);
    if (other != cur) {
      /* merge the POS tags in */
      tl1 = (treebank_label)cur->contents;
      tl2 = (treebank_label)other->contents;
      for (i=0; i<tl1->n_labels; i++) {
	found = 0;
	for (j=0; j<tl2->n_labels; j++)
	  if (!strcmp(tl1->labels[i], tl2->labels[j]))
	    found = 1;
	if (!found) {
	  tl2->labels = realloc(tl2->labels, (tl2->n_labels+1)*sizeof(char *));
	  tl2->labels[tl2->n_labels] = strdup(tl1->labels[i]);
	  tl2->n_labels++;
	}
      }
    } else {
      node_detach(cur);
    }
  } else {
    for (i=n-1; i>=0; i--)
      process(node_child(cur, i));
  }
}

int main (int argc, char *argv[]) {
  node root;
  int i = 0;
  FILE *fp;
  int c;
  char prefix[FILENAME_MAX] = "tig0";

  while ((c = getopt(argc, argv, "p:")) != EOF)
    switch (c) {
    case 'p': strncpy(prefix, optarg, FILENAME_MAX); break;
    }

  d = new_lexicon((COMPARE)preterm_compare, (HASH)preterm_hash);

  while ((root = read_tree(stdin, (FROM_STRING)treebank_nonterm_from_string, (FROM_STRING)treebank_term_from_string)) != NULL) {

    process(root);

    i++;
    if (i % 100 == 0)
      fprintf(stderr, "%d sentences processed.\n", i);

  }

  strncat(prefix, ".lex", FILENAME_MAX-strlen(prefix));
  fp = fopen(prefix, "w");
  lexicon_write(d, fp, (WRITE)preterm_write);
  fclose(fp);
}
