/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <string.h>

#include "node.h"
#include "treebank_label.h"

#include "tables.h"

#define DEFAULT "*"
#define LEFT "first"
#define RIGHT "last"
#define ALL "all"
#define NOT "not"
#define HLEFT "first-after-head"
#define HRIGHT "last-before-head"

static int initialized = 0;
static node head_table, arg_table;

int subsumes(treebank_label tl1, treebank_label tl2) {
  int i, j, found;

  if (strcmp(tl1->label, tl2->label) && strcmp(tl1->label, DEFAULT))
    return 0;

  /* Now check tags */

  for (i=0; i<tl1->n_tags; i++) {
    found = 0;
    for (j=0; j<tl2->n_tags; j++) {
      if (!strcmp(tl1->tags[i], tl2->tags[j])) {
	found = 1;
	break;
      }
    }
    if (!found)
      return 0;
  }
  return 1;
}

void init_tables(void) {
  FILE *fp;
  int i, j;

  fp = fopen("heads.txt", "r");
  head_table = read_tree(fp, (FROM_STRING)treebank_nonterm_from_string, (FROM_STRING)treebank_nonterm_from_string);
  fclose(fp);

  fp = fopen("args.txt", "r");
  arg_table = read_tree(fp, (FROM_STRING)treebank_nonterm_from_string, (FROM_STRING)treebank_nonterm_from_string);
  fclose(fp);

  initialized = 1;
}

int table_lookup(node table, node parent, int h, int *a) {
  int i, n;
  int n_entries, n_rules, n_children;
  int e, r, c;
  node entry, rule, child;
  const char *keyword;

  /* Start with blank slate */
  n_children = node_n_children(parent);
  for (c=0; c<n_children; c++)
    a[c] = 0;
  
  n_entries = node_n_children(table);
  for (e=0; e<n_entries; e++) {
    
    entry = node_child(table, e);
    if (subsumes(entry->contents, parent->contents)) {
      
      n_rules = node_n_children(entry);

      for (r=0; r<n_rules; r++) {
	rule = node_child(entry, r);
	keyword = ((treebank_label)rule->contents)->label;
      
	n = node_n_children(rule);

	if (!strcmp(keyword, ALL)) {
	  for (c=0; c<n_children; c++)
	    if (!a[c])
	      for (i=0; i<n; i++)
		if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		  a[c] = 1;
		  break;
		}
	} else if (!strcmp(keyword, NOT)) {
	  for (c=0; c<n_children; c++)
	    if (a[c])
	      for (i=0; i<n; i++)
		if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		  a[c] = 0;
		  break;
		}
	} else if (!strcmp(keyword, LEFT)) {
	  for (c=0; c<n_children; c++)
	    for (i=0; i<n; i++)
	      if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		a[c] = 1;
		return 0;
	      }
	} else if (!strcmp(keyword, RIGHT)) {
	  for (c=n_children-1; c>=0; c--)
	    for (i=0; i<n; i++)
	      if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		a[c] = 1;
		return 0;
	      }
	} else if (!strcmp(keyword, HLEFT)) {
	  for (c=h+1; c<n_children; c++)
	    for (i=0; i<n; i++)
	      if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		a[c] = 1;
		return 0;
	      }
	} else if (!strcmp(keyword, HRIGHT)) {
	  for (c=n_children-1; c>=h+1; c--)
	    for (i=0; i<n; i++)
	      if (subsumes(node_child(rule, i)->contents, node_child(parent, c)->contents)) {
		a[c] = 1;
		return 0;
	      }
	} else {
	  fprintf(stderr, "Warning: unsupported keyword %s\n", keyword);
	}
      }
    }
  }
  return 0;
}

int head_table_lookup(node parent) {
  int *a, i, n;
  if (!initialized)
    init_tables();

  n = node_n_children(parent);
  if (n == 1) {
    return 0;
  } else if (n > 0) {
    a = malloc(sizeof(int)*n);
    table_lookup(head_table, parent, 0, a);
    for (i=0; i<n; i++)
      if (a[i]) {
	free(a);
	return i;
      }
    free(a);
    return -1;
  } else 
    return -1;
}

int arg_table_lookup(node parent, int h, int *arg) {
  if (!initialized)
    init_tables();

  return table_lookup(arg_table, parent, h, arg);
}

