/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <string.h>
#include <stdlib.h>

#include "tag.h"

int multicomponent = 0;

elementary_tree new_elementary_tree(void) {
  elementary_tree et = malloc(sizeof(*et));
  et->root = NULL;

  et->index = 0;

  return et;
}

void elementary_tree_delete(elementary_tree et) {
  tree_delete(et->root, (DELETE)derived_label_delete);
  free(et);
}

void elementary_tree_copy_helper(elementary_tree et, node cur) {
  int i, n;
  ((derived_label)cur->contents)->elementary_tree = et;

  n = node_n_children(cur);
  for (i=0; i<n; i++)
    elementary_tree_copy_helper(et, node_child(cur, i));
}

elementary_tree elementary_tree_copy(elementary_tree et) {
  elementary_tree new = new_elementary_tree();
  new->type = et->type;
  new->root = tree_copy(et->root, (ENDO)derived_label_copy);
  elementary_tree_copy_helper(new, new->root);
  new->index = et->index;

  new->n_sites = et->n_sites;

  return new;
}

void elementary_tree_print(FILE *fp, elementary_tree et) {
  tree_print(fp, et->root, (TO_STRING)derived_label_to_string);
}

int max_type(node cur) {
  int i, n, max, cmax;

  max = ((derived_label)cur->contents)->type;
  n = node_n_children(cur);
  for (i=0; i<n; i++) {
    cmax = max_type(node_child(cur, i));
    if (cmax > max)
      max = cmax;
  }
  return max;
}

elementary_tree read_elementary_tree(FILE *fp) {
  elementary_tree et = new_elementary_tree();

  et->root = read_tree(fp, (FROM_STRING)derived_label_from_string, (FROM_STRING)derived_label_from_string);

  et->n_sites = max_type(et->root)+1;
  if (et->n_sites < 0)
    et->n_sites = 0;

  return et;
}

/* Compares two elementary trees without regard to their indexes. */

int elementary_tree_compare_helper (node n1, node n2) {
  int i, n;
  derived_label dl1, dl2;

  dl1 = ((derived_label)n1->contents);
  dl2 = ((derived_label)n2->contents);

  if (strcmp(dl1->label, dl2->label))
    return 1;

  /* This test ignores different classes of adjunction nodes. */

  if (!((dl1->type == -2 && dl2->type == -2) ||
	(dl1->type == -1 && dl2->type == -1) ||
	(dl1->type >= 0 && dl2->type >=0)))
    return 1;

  n = node_n_children(n1);
  if (n != node_n_children(n2))
    return 1;
  for (i=0; i<n; i++)
    if (elementary_tree_compare_helper(node_child(n1, i), node_child(n2, i)))
      return 1;
  return 0;
}

int elementary_tree_compare (elementary_tree et1, elementary_tree et2) {
/*  if (et1->type != et2->type)
    return 1;*/

  if (et1->root == NULL && et2->root == NULL)
    return 0;
  else if (et1->root == NULL || et2->root == NULL)
    return 1;

  return elementary_tree_compare_helper(et1->root, et2->root);
}


/* Not a very good hash function. */

unsigned int elementary_tree_hash_helper (node cur) {
  int i, n;
  unsigned int hash = 0;
  char *label = ((derived_label)cur->contents)->label;

  n = strlen(label);
  for (i=0; i<n; i++)
    hash = hash*2 + label[i];

  n = node_n_children(cur);
  for (i=0; i<n; i++)
    hash += elementary_tree_hash_helper(node_child(cur, i));

  return hash;
}

unsigned int elementary_tree_hash (elementary_tree et) {
  return elementary_tree_hash_helper(et->root);
}

derived_label new_derived_label(void) {
  derived_label dl = malloc(sizeof(*dl));
  dl->label = NULL;
  dl->type = -2;
  dl->elementary_tree = NULL;
  dl->anchor = 0;
  dl->head = 0;
  dl->trivials = 0;

  dl->imin = dl->imax = dl->jmin = dl->jmax = -1;

  return dl;
}

void derived_label_delete(derived_label dl) {
  free(dl->label);
  free(dl);
}

derived_label derived_label_copy(derived_label dl) {
  derived_label new = new_derived_label();
  new->label = strdup(dl->label);
  new->type = dl->type;
  new->elementary_tree = dl->elementary_tree;
  new->anchor = dl->anchor;
  new->spine = dl->spine;
  new->addr = dl->addr;
  new->head = dl->head;

  return new;
}

int derived_label_to_string(derived_label dl, char *s, size_t n) {
  char buf[100];

  strncpy(s, dl->label, n);

  if (dl->type == -1)
    strncat(s, "*", n-strlen(s));
  else if (dl->type >= 0) {
    if (multicomponent) {
      sprintf(buf, "+%d", dl->type);
      strncat(s, buf, n-strlen(s));
    } else {
      strncat(s, "+", n-strlen(s));
    }
  }

  return 0;
}

derived_label derived_label_from_string(const char *s) {
  derived_label dl = new_derived_label();
  int n;
  char *s1;
  
  n = strlen(s);
  dl->label = strdup(s);

  if (!strcmp(dl->label, "@"))
    dl->anchor = 1;

  if (s[n-1] == '*') {
    dl->label[n-1] = '\0';
    dl->type = TYPE_FOOT;
  } else {
    s1 = strrchr(dl->label, '+');
    if (s1) {
      *s1 = '\0';
      if (multicomponent) {
	dl->type = atoi(s1+1);
      } else {
	dl->type = 0; /* not exactly the right thing to do */
      }
    }
  }

  return dl;
}

derived_label derived_terminal_from_string(const char *s) {
  derived_label dl = new_derived_label();
  int n;
  
  n = strlen(s);
  dl->label = strdup(s);

  return dl;
}

int derived_label_compare(derived_label dl1, derived_label dl2) {
  return strcmp(dl1->label, dl2->label);
}

/* Finds the first node of the specified type. */

node find_site(node cur, int type) {
  int i, n;
  node result;

  if (((derived_label)cur->contents)->type == type)
    return cur;

  n = node_n_children(cur);
  for (i=0; i<n; i++) {
    result = find_site(node_child(cur, i), type);
    if (result != NULL)
      return result;
  }

  return NULL;
}

derivation_label new_derivation_label(void) {
  derivation_label dl = malloc(sizeof(*dl));
  dl->elementary_tree = NULL;
  dl->headword = dl->headtag = NULL;

  dl->optional = 0;

  dl->n_sites = dl->n_optional = 0;
  dl->max_optional = 0;
  dl->sites = NULL;
  dl->slots = NULL;

  return dl;
}

derivation_label derivation_label_copy(derivation_label dl) {
  derivation_label ndl;
  int i;

  if (dl == NULL)
    return NULL;

  ndl = new_derivation_label();
  ndl->elementary_tree = dl->elementary_tree;
  if (dl->headword)
    ndl->headword = strdup(dl->headword);
  if (dl->headtag)
    ndl->headtag = strdup(dl->headtag);

  ndl->optional = dl->optional;
  ndl->site = dl->site;
  ndl->slot = dl->slot;

  ndl->n_sites = dl->n_sites;
  ndl->n_optional = dl->n_optional;

  ndl->max_optional = dl->max_optional;

  if (dl->max_optional > 0) {
    ndl->sites = malloc(dl->max_optional * sizeof(node));
    ndl->slots = malloc(dl->max_optional * sizeof(int));
    
    for (i=0; i<dl->n_optional; i++) {
      ndl->sites[i] = dl->sites[i];
      ndl->slots[i] = dl->slots[i];
    }
  } else {
    ndl->sites = NULL;
    ndl->slots = NULL;
  }

  return ndl;
}

/* NB this function does not free elementary_tree. */

void derivation_label_delete(derivation_label dl) {
  free(dl->headword);
  free(dl->headtag);
  free(dl->sites);
  free(dl->slots);
  free(dl);
}

int derivation_label_to_string(derivation_label dl, char *s, size_t n) {
  static char buf[100];
  int i;

  if (dl == NULL)
    strncpy(s, "NIL", n);
  else if (dl->headword && dl->headtag)
    sprintf(s, "%d-%s-%s", dl->elementary_tree->index, dl->headword, dl->headtag);
  else if (dl->headword)
    sprintf(s, "%d-%s", dl->elementary_tree->index, dl->headword);
  else
    sprintf(s, "%d", dl->elementary_tree->index);

  if (dl->n_optional > 0) {
    for (i=0; i<dl->n_optional; i++) {
      sprintf(buf, ",@%s%d", ((derived_label)dl->sites[i]->contents)->label, dl->slots[i]);
      strcat(s, buf);
    }
  }

  return 0;
}

