/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#ifndef TAG_H
#define TAG_H

#include <stdio.h>

#include "node.h"
#include "list.h"

typedef struct {
  int type; /* initial or auxiliary */
  node root;
  int index;

  int n_sites;

  /* for accounting */
  int count; 
} *elementary_tree;

elementary_tree new_elementary_tree(void);
void elementary_tree_delete(elementary_tree et);
elementary_tree elementary_tree_copy(elementary_tree et);
void elementary_tree_print(FILE *fp, elementary_tree et);
elementary_tree read_elementary_tree(FILE *fp);

int elementary_tree_compare (elementary_tree et1, elementary_tree et2);
unsigned int elementary_tree_hash (elementary_tree et);


typedef struct {
  /* Basic information */
  char *label;
  int type; /* >=0 = adjunction node */
#define TYPE_FOOT -1
#define TYPE_OTHER -2

  elementary_tree elementary_tree;

  /* For extraction */
  int trivials;
  node percolated;

  /* For singly-lexicalized TAG */
  int anchor;

  /* For convenience */
  int spine, addr;

  /* For compatibility with current parser */
  int head;

  /* For parser */
  int imin, imax, jmin, jmax;

} *derived_label;

int derived_label_to_string(derived_label dl, char *s, size_t n);
derived_label derived_label_from_string(const char *s);
derived_label derived_terminal_from_string(const char *s);
derived_label new_derived_label(void);
void derived_label_delete(derived_label dl);
derived_label derived_label_copy(derived_label dl);
int derived_label_compare(derived_label dl1, derived_label dl2);

typedef struct {
  elementary_tree elementary_tree;
  char *headword, *headtag;

  /* These are used by the trainer. Should change to match scheme
     used by parser. */
  int optional;
  node site;
  int slot;

  /* These are used by the parser. */
  int n_sites, n_optional, max_optional;

  node *sites;
  int *slots;
} *derivation_label;

derivation_label new_derivation_label(void);
void derivation_label_delete(derivation_label dl);
int derivation_label_to_string(derivation_label dl, char *s, size_t n);
derivation_label derivation_label_copy(derivation_label dl);

#endif
