/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#include "generic.h"
#include "node.h"
#include "tag.h"
#include "lexicon.h"
#include "unknown.h"
#include "treebank_label.h"

void init_markup(void);
node find_heads(node cur);
node extract(node cur, node deriv);
int count_events(node deriv, const char *code, lexicon words, int threshold);

int verbose = 1;
int delexicalize = 2, binarize = 1, use_code = 1, left_to_right = 1;
int predicatives = 1;
int code_level = 2; /* 0=none, 1=first, 2=label, 3=node */
int show_events = 1, show_growth = 0, show_derivations = 0;
int word_threshold = 6;
int read_existing_grammar = 0;

/* int weigh(node cur) {
  int i, n, weight;
  int preterm;
  char *label;

  weight = 0;

  n = node_n_children(cur);

  preterm = (n == 1 && node_n_children(node_child(cur, 0)) == 0);

  for (i=0; i<n; i++)
    weight += weigh(node_child(cur, i));

  if (preterm && ((treebank_label)cur->contents)->label[0] == 'V')
    weight++;

  if (weight > 2)
    weight = 2;

  if (n == 0 || preterm)
    ((treebank_label)cur->contents)->weight = -1;
  else
    ((treebank_label)cur->contents)->weight = weight;

  return weight;
} */

/* Assigns each elementary tree in deriv an index, and adds any new 
   trees to g. */

void register_trees(node deriv, lexicon g) {
  int i, n;
  elementary_tree et = ((derivation_label)deriv->contents)->elementary_tree;

  if (!tree_trivial(et)) {
    lexicon_insert(g, et);
    et->index = lexicon_id(g, et);
  }

  n = node_n_children(deriv);
  for (i=0; i<n; i++)
    register_trees(node_child(deriv, i), g);
}

/* Deletes all trees that didn't get incorporated into g. */

void cleanup_trees(node cur, lexicon g) {
  int i, n;
  elementary_tree et = ((derivation_label)cur->contents)->elementary_tree;

  /* First check to see if the tree is in g */
  if (lexicon_lookup(g, et) != et)
    elementary_tree_delete(et);

  n = node_n_children(cur);

  for (i=n-1; i>=0; i--)
    cleanup_trees(node_child(cur, i), g);

  derivation_label_delete((derivation_label)cur->contents);
  node_delete(cur);
}

/* For compatibility with the current parser */

node binarize_tree(node cur) {
  node new, child;
  int i, n;
  char *label;

/*  label = malloc(strlen(((derived_label)cur->contents)->label)+2);
  strcpy(label, ((derived_label)cur->contents)->label);
  strcat(label, "<");*/
  label = strdup("<");

  n = node_n_children(cur);

  if (n == 0)
    return cur;

  if (n == 1 && delexicalize < 2 && ((derived_label)node_child(cur, 0)->contents)->anchor)
    return cur;

  if (n == 1 && ((derived_label)cur->contents)->type >= 0) {
    child = node_child(cur, 0);
    node_supplant(cur, child);
    node_delete(cur);
    return binarize_tree(child);
  }

  if (((derived_label)cur->contents)->anchor) {
    if (delexicalize >= 2)
      node_delete(node_child(cur, 0));
    return cur;
  }

  for (i=0; i<n; i++)
    binarize_tree(node_child(cur, i));

  child = new = new_node(new_derived_label());
  ((derived_label)new->contents)->label = label;

  i = 0;
  while (i < n && ((derived_label)node_child(cur, 0)->contents)->head == 0) {
    node_insert_child(new, 0, node_child(cur, 0));
    node_insert_child(new, 1, new_node(new_derived_label()));
    new = node_child(new, 1);
    ((derived_label)new->contents)->label = strdup(label);
    i++;
  }

  node_insert_child(new, 0, new_node(new_derived_label()));
  new = node_child(new, 0);

/*  ((derived_label)new->contents)->label = strdup(label);
  ((derived_label)new->contents)->label[strlen(label)-1] = '>';*/
  label = strdup(">");
  ((derived_label)new->contents)->label = label;

  n = node_n_children(cur);
  i = n-1;
  while (i > 0) {
    node_insert_child(new, 0, new_node(new_derived_label()));
    node_insert_child(new, 1, node_child(cur, i));
    new = node_child(new, 0);
    ((derived_label)new->contents)->label = strdup(label);
    i--;
  }

  node_insert_child(new, 0, node_child(cur, 0));
  node_insert_child(cur, 0, child);

  return cur;
}

/* Mark as "head" every node on path from root to anchor.
   This gets done during extraction process, but we redo it
   for the case when a grammar is read from a file. */
void mark_heads(node cur) {
  derived_label dl = (derived_label)cur->contents, cdl;
  int n, i;

  n = node_n_children(cur);
  if (n == 0 && dl->anchor)
    dl->head = 1;
  for (i=0; i<n; i++) {
    mark_heads(node_child(cur, i));
    cdl = (derived_label)node_child(cur, i)->contents;
    if (cdl->head)
      dl->head = 1;
  }
}

void binarize_trees(lexicon g) {
  lexiconIterator li;
  elementary_tree et;

  li = new_lexiconIterator(g);
  while (lexiconIterator_hasNext(li)) {
    et = lexiconIterator_next(li);
    mark_heads(et->root);
    et->root = binarize_tree(et->root);
  }
}

void print_trees(node cur) {
  int i, n;

  printf("%d ", ((derivation_label)cur->contents)->elementary_tree->index);
  elementary_tree_print(stdout, ((derivation_label)cur->contents)->elementary_tree);
  printf("\n");

  n = node_n_children(cur);

  for (i=0; i<n; i++)
    print_trees(node_child(cur, i));
}


int main (int argc, char *argv[]) {
  node root, deriv;
  lexicon g, words;
  int i = 0;
  int c;
  FILE *fp;
  char prefix[FILENAME_MAX] = "tig0";
  char filename[FILENAME_MAX];

  while ((c = getopt(argc, argv, "hs:p:Ww:xgrb")) != EOF)
    switch (c) {
    case 'h': left_to_right = 0; break;
    case 's': code_level = atoi(optarg); break;
    case 'p': strncpy(prefix, optarg, FILENAME_MAX); break;
    case 'w': word_threshold = atoi(optarg); break;
    case 'W': word_threshold = -1; break;
    case 'x': predicatives = 0; break;
    case 'g': show_events = 0; show_growth = 1; break;
    case 'r': read_existing_grammar = 1; break;
    case 'b': binarize = 0; break;
    }

  words = new_lexicon((COMPARE)preterm_compare, (HASH)preterm_hash);
  sprintf(filename, "%s.lex", prefix);
  fp = fopen(filename, "r");
  lexicon_read(words, fp, (READ)preterm_read);
  fclose(fp);

  g = new_lexicon((COMPARE)elementary_tree_compare, (HASH)elementary_tree_hash);
  if (read_existing_grammar) {
    sprintf(filename, "%s.grm", prefix);
    fp = fopen(filename, "r");
    lexicon_read(g, fp, (READ)read_elementary_tree);
    fclose(fp);
  }

  while ((root = read_tree(stdin, (FROM_STRING)treebank_nonterm_from_string, (FROM_STRING)treebank_term_from_string)) != NULL) {

/*    raise_punct(root);*/

/*    weigh(root);*/

    deriv = node_root(find_heads(root));


    extract(root, deriv);

    register_trees(deriv, g);

    if (show_derivations) {
      print_trees(deriv);
      tree_print(stdout, deriv, (TO_STRING)derivation_label_to_string);
      printf("\n");
    }

    if (show_events)
      count_events(deriv, "NIL", words, word_threshold);

    cleanup_trees(deriv, g);

    i++;

    if (verbose  && i % 100 == 0) 
      fprintf(stderr, "%6d sentences processed.\n", i);

    if (show_growth)
      printf("%d %d\n", i, lexicon_size(g));

  }

  /* BUG: this destroys the hash table */
  if (binarize)
    binarize_trees(g);

  sprintf(filename, "%s.grm", prefix);
  fp = fopen(filename, "w");
  lexicon_write(g, fp, (WRITE)elementary_tree_print);
  fclose(fp);

  return 0;
}
