/* Copyright (c) 2001 by David Chiang. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if 0
#include <getopt.h>
#endif

#include "node.h"

int tagged = 0;

int tree_rules_print(FILE *fp, node cur, void (*to_string)(const void *, char *, int)) {
  int n, i;
  char buf[255];

  n = node_n_children(cur);

  if (n > 0) {
    (*to_string)(cur->contents, buf, sizeof(buf));
    fprintf(fp, "%s ->", buf);
    
    for (i=0; i<n; i++) {
      (*to_string)(node_child(cur, i)->contents, buf, sizeof(buf));
      fprintf(fp, " %s", buf);
    }
    
    fprintf(fp, "\n");
    
    for (i=0; i<n; i++)
      tree_rules_print(fp, node_child(cur, i), to_string);
  }
}

int tree_frontier_print(FILE *fp, node cur, void (*to_string)(const void *, char *, int), int leftmost) {
  int n, i;
  char buf[255], buf1[255];
  n = node_n_children(cur);
  if (n == 0) {
    if (!leftmost)
      putc(' ', fp);

    if (tagged) {
      (*to_string)(cur->contents, buf, sizeof(buf));
      if (cur->parent)
	(*to_string)(cur->parent->contents, buf1, sizeof(buf1));
      else
	strcpy(buf1, "?");
      fprintf(fp, "(%s %s)", buf1, buf);
    } else {
      (*to_string)(cur->contents, buf, sizeof(buf));
      fputs(buf, fp);
    }

  } else if (n > 0) {
    for (i=0; i<n; i++)
      tree_frontier_print(fp, node_child(cur, i), to_string, leftmost && i==0);
  }

  return 0;
}

int main (int argc, char *argv[]) {
  node cur;
  int c;
  int pretty = 0, frontier = 0, rules = 0;

  while ((c = getopt(argc, argv, "prft")) != EOF)
    switch(c) {
    case 'p': pretty = 1; break;
    case 'r': rules = 1; break;
    case 'f': frontier = 1; break;
    case 't': frontier = tagged = 1; break;
    }

  while (!feof(stdin)) {
    cur = read_tree(stdin, (FROM_STRING)string_from_string, (FROM_STRING)string_from_string);

    if (cur) {
      if (pretty)
	tree_pretty_print(stdout, cur, (TO_STRING)string_to_string);
      else if (rules)
	tree_rules_print(stdout, cur, (TO_STRING)string_to_string);
      else if (frontier)
	tree_frontier_print(stdout, cur, (TO_STRING)string_to_string, 1);
      else
	tree_print(stdout, cur, (TO_STRING)string_to_string);
      putchar('\n');
    }
  }

  return 0;
}

